#include <stdio.h>
#include <stdlib.h>
#ifdef _WIN32_
#include <io.h>
#else
#include <unistd.h>
#endif
#include <fcntl.h>

#include "dib.h"

int DIBLoad(char* fname, unsigned char* buf,
			int width, int height, int raw, int ox, int oy,
			unsigned char* rgb)
{
	unsigned long l, m;
	FILE* fp;
	int i, x, y, n;
	unsigned long bfOffBits, biSize, bfSize;
	int biWidth, biHeight, biBitCount;
	unsigned char uc;
	int f_raw;
	unsigned char* bits;
	
	if ((fp=fopen(fname, "rb"))==NULL) {
		return -1;
	}

	// BITMAPFILEHEADER.bfType(WORD)
	if (fgetc(fp)!='B') {
		fclose(fp);
		return -1;  // format error
	}
	if (fgetc(fp)!='M') {
		fclose(fp);
		return -1;  // format error
	}

	// BITMAPFILEHEADER.bfSize(DWORD)
	l = m = 0;
	for(i=0; i<4; i++) {
		m = (unsigned long)fgetc(fp);
		m <<= i*8;
		l |= m;
	}
//	printf("bfSize = %d\n", l);
	bfSize = l;

	// BITMAPFILEHEADER.bfReserved*(WORD*2)
	for(i=0; i<4; i++) {
		fgetc(fp);
	}

	// BITMAPFILEHEADER.bfOffBits(DWORD)
	l = m = 0;
	for(i=0; i<4; i++) {
		m = (unsigned long)fgetc(fp);
		m <<= i*8;
		l |= m;
	}
//	printf("bfOffBits = %d\n", l);
	bfOffBits = l;

	// BITMAPINFOHEADER.biSize
	l = m = 0;
	for(i=0; i<4; i++) {
		m = (unsigned long)fgetc(fp);
		m <<= i*8;
		l |= m;
	}
//	printf("biSize = %d\n", l);
	biSize = l;
	
	// BITMAPINFOHEADER.biWidth
	l = m = 0;
	for(i=0; i<4; i++) {
		m = (unsigned long)fgetc(fp);
		m <<= i*8;
		l |= m;
	}
//	printf("biWidth = %d\n", l);
	biWidth = l;
	
	// BITMAPINFOHEADER.biHeight
	l = m = 0;
	for(i=0; i<4; i++) {
		m = (unsigned long)fgetc(fp);
		m <<= i*8;
		l |= m;
	}
//	printf("biHeight = %d\n", l);
	biHeight = l;
	
	// BITMAPINFOHEADER.biPlanes(WORD)
	for(i=0; i<2; i++) {
		fgetc(fp);
	}
	
	// BITMAPINFOHEADER.biPlanes(WORD)
	l = m = 0;
	for(i=0; i<2; i++) {
		m = (unsigned long)fgetc(fp);
		m <<= i*8;
		l |= m;
	}
//	printf("biBitCount = %d\n", l);
	biBitCount = l;
	
	// ʲBITMAPINFOHEADERɤФ
	for(i=0; i<24; i++) {
		fgetc(fp);
	}

	// ʲѥåΰ
	if (biBitCount <= 8) {
		n = (1<<biBitCount);
		for(i=0; i<n; i++) {
			uc = (unsigned char)fgetc(fp); // blue
			if (rgb) rgb[i*3 + 2] = uc;
			uc = (unsigned char)fgetc(fp); // green
			if (rgb) rgb[i*3 + 1] = uc;
			uc = (unsigned char)fgetc(fp); // red
			if (rgb) rgb[i*3 + 0] = uc;
			fgetc(fp);  // ɤФ
		}
	}

	// ӥåȥ᡼ɤ߹
	fseek(fp, bfOffBits, SEEK_SET);
	f_raw = (((biWidth*biBitCount + 7) / 8) + 3)/4*4;
	bits = (unsigned char*)malloc(f_raw*biHeight);
	for(i=0; i<f_raw*biHeight; i++) bits[i] = 0;
	fread(bits, f_raw*biHeight, 1, fp);
	fclose(fp);

	if (biBitCount==4) {
		for(y=0; y<height; y++) {
			for(x=0; x<width; x++) {
				if (x%2==0) {
					if (oy + y < biHeight && ox+x/2 < biWidth) {
						buf[y*raw + x] = bits[(oy+y)*f_raw + (ox+x/2)]>>4;
					} else {
						buf[y*raw + x] = 0;
					}
				} else {
					if (oy + y < biHeight && ox+x/2 < biWidth) {
						buf[y*raw + x] = bits[(oy+y)*f_raw + (ox+x/2)] & 0x0f;
					} else {
						buf[y*raw + x] = 0;
					}
				}
			}
		}
	}
	if (biBitCount==8) {
		for(y=0; y<height; y++) {
			for(x=0; x<width; x++) {
				if (oy + y < biHeight && ox + x < biWidth) {
					buf[y*raw + x] = bits[(oy+y)*f_raw + (ox+x)];
				} else {
					buf[y*raw + x] = 0;
				}
			}
		}
	}
	if (biBitCount==24) {
		for(y=0; y<height; y++) {
			for(x=0; x<width; x++) {
				if (oy+y < biHeight && ox+x < biWidth) {
					buf[y*raw + x*3 + 0] = bits[(oy+y)*f_raw + (ox+x)*3 + 0];
					buf[y*raw + x*3 + 1] = bits[(oy+y)*f_raw + (ox+x)*3 + 1];
					buf[y*raw + x*3 + 2] = bits[(oy+y)*f_raw + (ox+x)*3 + 2];
				} else {
					buf[y*raw + x*3 + 0] = 0;
					buf[y*raw + x*3 + 1] = 0;
					buf[y*raw + x*3 + 2] = 0;
				}
			}
		}
	}
	free(bits);
	return 0;
}



int DIBInfo(char* fname, int* width, int* height, int* depth,
			unsigned char* rgb)
{
	unsigned long l, m;
	FILE* fp;
	int i, n;
	unsigned long bfOffBits, biSize, bfSize;
	int biWidth, biHeight, biBitCount;
	unsigned char uc;
	
	if ((fp=fopen(fname, "rb"))==NULL) {
		return -1;
	}

	// BITMAPFILEHEADER.bfType(WORD)
	if (fgetc(fp)!='B') {
		fclose(fp);
		return -1;  // format error
	}
	if (fgetc(fp)!='M') {
		fclose(fp);
		return -1;  // format error
	}

	// BITMAPFILEHEADER.bfSize(DWORD)
	l = m = 0;
	for(i=0; i<4; i++) {
		m = (unsigned long)fgetc(fp);
		m <<= i*8;
		l |= m;
	}
//	printf("bfSize = %d\n", l);
	bfSize = l;

	// BITMAPFILEHEADER.bfReserved*(WORD*2)
	for(i=0; i<4; i++) {
		fgetc(fp);
	}

	// BITMAPFILEHEADER.bfOffBits(DWORD)
	l = m = 0;
	for(i=0; i<4; i++) {
		m = (unsigned long)fgetc(fp);
		m <<= i*8;
		l |= m;
	}
//	printf("bfOffBits = %d\n", l);
	bfOffBits = l;

	// BITMAPINFOHEADER.biSize
	l = m = 0;
	for(i=0; i<4; i++) {
		m = (unsigned long)fgetc(fp);
		m <<= i*8;
		l |= m;
	}
//	printf("biSize = %d\n", l);
	biSize = l;
	
	// BITMAPINFOHEADER.biWidth
	l = m = 0;
	for(i=0; i<4; i++) {
		m = (unsigned long)fgetc(fp);
		m <<= i*8;
		l |= m;
	}
//	printf("biWidth = %d\n", l);
	biWidth = l;
	
	// BITMAPINFOHEADER.biHeight
	l = m = 0;
	for(i=0; i<4; i++) {
		m = (unsigned long)fgetc(fp);
		m <<= i*8;
		l |= m;
	}
//	printf("biHeight = %d\n", l);
	biHeight = l;
	
	// BITMAPINFOHEADER.biPlanes(WORD)
	for(i=0; i<2; i++) {
		fgetc(fp);
	}
	
	// BITMAPINFOHEADER.biPlanes(WORD)
	l = m = 0;
	for(i=0; i<2; i++) {
		m = (unsigned long)fgetc(fp);
		m <<= i*8;
		l |= m;
	}
//	printf("biBitCount = %d\n", l);
	biBitCount = l;
	
	// ʲBITMAPINFOHEADERɤФ
	for(i=0; i<24; i++) {
		fgetc(fp);
	}

	// ʲѥåΰ
	if (biBitCount <= 8) {
		n = (1<<biBitCount);
		for(i=0; i<n; i++) {
			uc = (unsigned char)fgetc(fp); // blue
			if (rgb) rgb[i*3 + 2] = uc;
			uc = (unsigned char)fgetc(fp); // green
			if (rgb) rgb[i*3 + 1] = uc;
			uc = (unsigned char)fgetc(fp); // red
			if (rgb) rgb[i*3 + 0] = uc;
			fgetc(fp);  // ɤФ
		}
	}
	fclose(fp);

	*width = biWidth;
	*height = biHeight;
	*depth = biBitCount;
	return 0;
}



// 8bit
int DIBSave(char* fname, unsigned char* buf,
			int width, int height, int raw, unsigned char* rgb)
{
	return 0;
}



